!====================================================================
! file bt_drv.f90
!====================================================================
!         EFFECTIVE METHODS FOR SOLVING BANDED TOEPLITZ SYSTEMS
!                     by D.A. Bini and B. Meini
!               (bini@dm.unipi.it    meini@dm.unipi.it)
!           Fortran 90 Program version 1.0, October 30, 1997
!--------------------------------------------------------------------
!                   S C A L A R        C A S E
!====================================================================
!                           DRIVER PROGRAM
!====================================================================

!====================================================================
! This is the driver program for solving the banded Toeplitz system
! Tx=b by means of cyclic reduction. The matrix T is defined by the
! scalar  numbers a_i, where a_i=0 for |i|>m, such that T=(a_{i-j}).
! The matrix T is reblocked in the tridiagonal form
!
!        |  A_0  A_{-1}                          |
!        |  A_1    A_0   A_{-1}                  |
!        |         A_1    A_0   A_{-1}           |    
! T =    |            ...   ...   ...            |
!        |                   ...   ...           |
!        |                         ...    A_{-1} |
!        |                         A_1    A_0    |
!
! where A_i have dimension m. It is supposed that both m and the
! block dimension nb of T are powers of 2.
!====================================================================
! The program reads from the standard input the name of the file
! (<input_file_name>) containing the data. Then reads from the file
! <input_file_name> the data and calls the subroutine "solve_cr"
! solving the system.      
! The computed  solution is stored in the file <input_file_name.out> 
! The file <input_file_name> must contain on each row the following
! data: 
! 1-st row: EPS ( real(kind(0.d0)) ) = Precision level for cyclic
! reduction 
! 2-nd row: EPSCG ( real(kind(0.d0)) = Precision level for the
! conjugate gradient method (used for the inversion of the diagonal
! Toeplitz-like blocks)
! 3-rd row: NB (integer) = The block dimension of the system
! 4-th row: M  (integer) = The dimension of the blocks
! The subsequent rows: the values
! a_{-M}
! a_{-M+1}
!  :
! a_{M-1}
! a_M
! b_1
! b_2
!  :
! b_{M*NB}
!====================================================================
PROGRAM driver
  USE bt_interface
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER   :: a,  b, x
  REAL(KIND(0.d0))                            :: eps, epsgc
  INTEGER                                     :: m, nb, i, j, ii
  CHARACTER(len=10)                           :: ifn
  LOGICAL                                     :: ex
  WRITE(*,*)'-------------------------------------------------------'
  WRITE(*,*)' EFFECTIVE METHODS FOR SOLVING BANDED TOEPLITZ SYSTEMS'
  WRITE(*,*)'                 by D.A. Bini and B. Meini'
  WRITE(*,*)'   Fortran 90 Program version 1.0, October 30, 1997'    
  WRITE(*,*)'-------------------------------------------------------'
  WRITE(*,*)
  WRITE(*,'(A)', advance='NO')' Input-file name = '
  READ(*,*) ifn
  INQUIRE(file=ifn, exist=ex)
  IF(.NOT.ex)THEN
     WRITE(*,*)
     WRITE(*,*)'THE FILE      ',ifn,'DOES NOT EXIST'
     STOP
  ENDIF
  OPEN(unit=10,file=ifn,status='old')
  J = INDEX(ifn,'.')
  IF (J.EQ.0) J = INDEX(ifn,' ')
  OPEN(unit=11,file=ifn(1:j-1)//'.out')
  READ(10,*) eps
  READ(10,*) epsgc
  READ(10,*) nb
  READ(10,*) m
  ALLOCATE(a(m,3))
  ALLOCATE(x(m,nb))
  ALLOCATE(b(m,nb))
  a=0.d0
  b=0.d0
  ! reading data
  DO i=1,m
     READ(10,*)a(i,1)
  END DO
  DO i=1,m
     READ(10,*)a(i,2)
  END DO
  READ(10,*)a(1,3)
  DO j=1,nb
     DO i=1,m
        READ(10,*)b(i,j)
     END DO
  END DO
  CALL solve_cr(a,b,x,eps,epsgc)
  WRITE(11,*)'-------------------------------------------------------'
  WRITE(11,*)' EFFECTIVE METHODS FOR SOLVING BANDED TOEPLITZ SYSTEMS'
  WRITE(11,*)'                 by D.A. Bini and B. Meini'
  WRITE(11,*)'   Fortran 90 Program version 1.0, October 30, 1997'    
  WRITE(11,*)'-------------------------------------------------------'
  WRITE(11,*)
  WRITE(11,*)"x="
  DO i=1,nb
     ii=(i-1)*m
     DO j=1,m
        WRITE(11,*)ii+j,x(j,i)
     END DO
  END DO
END PROGRAM driver



















