!====================================================================
! file bt_sub.f90
!====================================================================
!         EFFECTIVE METHODS FOR SOLVING BANDED TOEPLITZ SYSTEMS
!                     by D.A. Bini and B. Meini
!               (bini@dm.unipi.it    meini@dm.unipi.it)
!           Fortran 90 Program version 1.0, March 30, 1997
!--------------------------------------------------------------------
!                   S C A L A R        C A S E
!====================================================================
!                          SUBROUTINES
!====================================================================

!====================================================================
!                          SUBROUTINE SOLVE_CR
!====================================================================
! This subroutine solves the system Ax=b by means of cyclic
! reduction. The matrix A is banded Toeplitz. 
!====================================================================
! Input variables:
!    a  : pointer associated with the first block column of the
!         matrices A_i, i=-1,0,1.
!    b  : pointer associated with the vector b
!   eps : error bound for the stop condition
! epsgc : error bound for the CG method
!====================================================================
! Output variables: the solution x
!====================================================================
SUBROUTINE solve_cr(a,b,x,eps,epsgc)
  USE bt_interface, ONLY : fft_cr, bt_bcr, test, store, recover
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: a, x, b
  REAL(KIND(0.d0))                                :: eps, epsgc, err
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       ::  v
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: t, dft, bev
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: thb, ta, tr,&
       & bm0, r, ha
  LOGICAL                                         :: conv
  INTEGER                                         :: m, n, iter, nb,ln
  INTRINSIC size
  m=SIZE(a,1)
  n=m+m
  nb=SIZE(b,2)
  ALLOCATE(ha(m),v(m,2),r(m),t(n,5),dft(n,6),ta(n),thb(n),tr(n))
  ALLOCATE(bm0(m),bev(m,nb/2))
  bm0=a(m:1:-1,1)
  v=0.d0
  v(1,2)=1.d0
  r(:)=a(:,2)
  ha=a(:,2)
  err=eps+1
  iter=0
  ln=LOG(nb*1.d0)/LOG(2.d0)
  IF(2**ln<nb)ln=ln+1
  cr: DO WHILE((err>eps).AND.(iter<ln))
     iter=iter+1
     CALL fft_cr(a,v,r,dft,tr,ta,thb,t,epsgc) 
     CALL bt_bcr(a,v,ha,r,b,bev,dft,tr,ta,thb,t)
     CALL store(dft,tr,ta,thb,bev,t)
     CALL test(a,v,r,err)
!     WRITE(*,*)"Step=", iter, "Norm of the subdiagonal block=",err
  END DO cr
  CALL fft_cr(a,v,r,dft,tr,ta,thb,t,epsgc) 
  conv= (iter /= ln)
  CALL recover(conv,bm0,ha,r(1),dft,tr,ta,thb,t,b,x,epsgc) 
END SUBROUTINE solve_cr


!====================================================================
!                          SUBROUTINE BT_CR
!====================================================================
! This subroutine performs one step of cyclic reduction by computing
! the block rows and block columns defining the blocks A_i, i=-1,0,1
! and  \widehat A_1.
!====================================================================
! Input variables:
!    a  : pointer associated with the first block column of the
!         matrices A_i, i=-1,0,1.
!    v  : pointer associated with the displacement structure of the
!         matrices A_i, i=-1,0,1, \widehat A_1.
!    ha : pointer associated with the displacement structure of the
!         matrix \widehat A_1.
!    r  : pointer associated with the displacement structure of the
!         matrices A_1, \widehat A_1.
!   eps : error bound for the CG method
!====================================================================
! Output variables: new values for a, v, ha, r; 
!====================================================================
SUBROUTINE bt_cr(a,v,ha,r,err,eps)
  USE comtr
  USE tdft_interface
  USE fft_interface
  USE bt_interface, ONLY : acr, vcr, btl_inv, test, a0invv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: a, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: ha, r
  REAL(KIND(0.d0))                                :: err, eps
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER, SAVE :: t, dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE   :: thb, ta, tr, y 
  INTEGER                                         :: m, n
  INTRINSIC size
  m=SIZE(a,1)
  n=m+m
  IF(.NOT.ASSOCIATED(ta))THEN
     ALLOCATE(t(n,5))
     ALLOCATE(dft(n,6))
     ALLOCATE(ta(n))
     ALLOCATE(thb(n))
     ALLOCATE(tr(n))
     ALLOCATE(y(m))
  ELSE
     IF(n/=SIZE(ta,1))THEN
        NULLIFY(t)
        NULLIFY(dft)
        NULLIFY(ta)
        NULLIFY(thb)
        NULLIFY(tr)
        NULLIFY(y)
        ALLOCATE(t(n,5))
        ALLOCATE(dft(n,6))
        ALLOCATE(ta(n))
        ALLOCATE(thb(n))
        ALLOCATE(tr(n))
        ALLOCATE(y(m))
     ENDIF
  ENDIF
  y=a(:,1)
  CALL ldft(y,ta)
  dft(:,1)=ta           
  CALL l_uzj(y(1),ta,tr)
  dft(:,6)=tr    
  y=v(:,1)
  CALL ldft(y,ta)
  dft(:,2)=ta    
  CALL l_uzj(y(1),ta,tr)
  dft(:,4)=-tr           
  dft(1:m+1,4)=dft(1:m+1,4)+1.d0
  y=v(:,2)
  CALL ldft(y,ta)  
  dft(:,3)=ta         
  CALL l_uzj(y(1),ta,tr)
  dft(:,5)=tr         
  CALL ldft(r,tr)     
  CALL l_uzj(r(1),tr,thb)
  y = a(:,3)
  CALL ldft(y,ta)  
  CALL btl_inv(dft,v,a,t,eps)
  CALL vcr(v,t,dft,thb,tr,ta)
  CALL acr(a,ha,r,t,dft,thb,tr,ta)
  CALL test(a,v,r,err)
END SUBROUTINE bt_cr

!====================================================================
!                          SUBROUTINE  FFT_CR
!====================================================================
! This subroutine computes the DFT associated with the vectors a,v,r
! and with the matrix A_0^{-1}
!====================================================================
! Input variables:
!    a  : pointer associated with the first block column of the
!         matrices A_i, i=-1,0,1.
!    v  : pointer associated with the displacement structure of the
!         matrices A_i, i=-1,0,1, \widehat A_1.
!    r  : pointer associated with the displacement structure of the
!         matrices A_1, \widehat A_1.
!   eps : error bound for the CG method
!====================================================================
! Output variables: dft, tr, ta, thb, t
!====================================================================
SUBROUTINE fft_cr(a,v,r,dft,tr,ta,thb,t,eps)
  USE tdft_interface  
  USE comtr
  USE fft_interface
  USE bt_interface, ONLY : btl_inv
  IMPLICIT NONE
  REAL(KIND(0.d0))                                :: eps
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: a, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: r 
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: t, dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: thb, ta, tr
  ! auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE   :: y
  INTEGER                                         :: m, n
  INTRINSIC size
  m=SIZE(a,1)
  n=m+m
  IF(m/=SIZE(y,1))NULLIFY(y)
  IF(.NOT.ASSOCIATED(y))ALLOCATE(y(m))
  y=a(:,1)
  CALL ldft(y,ta)
  dft(:,1)=ta 
  CALL l_uzj(y(1),ta,tr)
  dft(:,6)=tr 
  y=v(:,1)
  CALL ldft(y,ta)
  dft(:,2)=ta 
  CALL l_uzj(y(1),ta,tr)
  dft(:,4)=-tr           
  dft(1:m+1,4)=dft(1:m+1,4)+1.d0 
  y=v(:,2)
  CALL ldft(y,ta)  
  dft(:,3)=ta           
  CALL l_uzj(y(1),ta,tr)
  dft(:,5)=tr           
  CALL ldft(r,tr)       
  CALL l_uzj(r(1),tr,thb)
  y = a(:,3)
  CALL ldft(y,ta)        
  CALL btl_inv(dft,v,a,t,eps)
END SUBROUTINE fft_cr


!====================================================================
!                          SUBROUTINE BT_BCR
!====================================================================
! This subroutine performs one step of cyclic reduction by computing
! the block rows and block columns defining the blocks A_i, i=-1,0,1
! and  \widehat A_1, and the new vector b
!====================================================================
! Input variables:
!    a  : pointer associated with the first block column of the
!         matrices A_i, i=-1,0,1.
!    v  : pointer associated with the displacement structure of the
!         matrices A_i, i=-1,0,1, \widehat A_1.
!    ha : pointer associated with the displacement structure of the
!         matrix \widehat A_1.
!    r  : pointer associated with the displacement structure of the
!         matrices A_1, \widehat A_1.
!    b  : pointer associated with the vector b
!   dft, tr, ta, thb, t : pointers associated with the DFT of
! A_i, \widehat A_1, A_0^{-1}
!====================================================================
! Output variables: new values for a, v, ha, r, b; 
! bev: pointer associated with T^{-1}b_+
!====================================================================
SUBROUTINE bt_bcr(a,v,ha,r,b,bev,dft,tr,ta,thb,t)
  USE bt_interface, ONLY : acr, vcr, bcr
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: a, v, b, bev
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: ha, r
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: t, dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: thb, ta, tr
  CALL vcr(v,t,dft,thb,tr,ta)
  CALL acr(a,ha,r,t,dft,thb,tr,ta)
  CALL bcr(b,bev,t,dft,thb,tr,ta)
END SUBROUTINE bt_bcr


!====================================================================
!                          SUBROUTINE BTL_INV
!====================================================================
! This subroutine computes the DFT associated with A_0^{-1}
!====================================================================
! Input variables:
!  dft  : pointer associated with the DFT of A_0
!    v  : pointer associated with the displacement structure of the
!         matrices A_i, i=-1,0,1, \widehat A_1.
!    a  : pointer associated with the first column of A_i
!  eps  : error bound for the CG method
!====================================================================
! Output variables: t, pointer associated with the DFT of A_0^{-1}.
! More specifically: 
! - t(1) is the DFT associated with the Toeplitz matrix L(..)
! + L^T(Z...)
! - t(2) is the DFT associated with L(A_0^{-1}a_{-1});
! - t(3) is the DFT associated with L^T(Z A_0^{-T}u_0);
! - t(4) is the DFT associated with L(A_0^{-1}v_0); 
! - t(5) is the DFT associated with L^T(Z A_0^{-t}b_{-1}).
!====================================================================
SUBROUTINE btl_inv(dft,v,a,t,eps)
  USE comtr
  USE tdft_interface
  USE fft_interface
  USE bt_interface, ONLY : ax, atx, cgata, cgaat
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER       :: dft, v, a, t
  REAL(KIND(0.d0))                                :: eps
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER, SAVE :: tdft, s
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE   :: ta0, tb0, w, x
  REAL(KIND(0.d0)), DIMENSION(:), POINTER         :: z, y
  INTEGER                                         :: m, n, i
  INTRINSIC size
  m=SIZE(v,1)
  n=m+m
  IF(.NOT.ASSOCIATED(w))THEN
     ALLOCATE(s(m,4),tdft(n,8),ta0(n),tb0(n),w(n),uv(n),lzu(n),x(m))
     s=0.d0
  ELSE
     IF(n/=SIZE(w,1))THEN
        NULLIFY(s,tdft,ta0,tb0,w,x,uv,lzu)
        ALLOCATE(s(m,4),tdft(n,8),ta0(n),tb0(n),uv(n),lzu(n),w(n)&
             & ,x(m))
        s=0.d0
     ENDIF
  ENDIF
  y => dft(:,2)
  CALL l_lzj_u(v(1,1),y,ta0,tb0)
  tdft(:,1)=-ta0              
  tdft(1:m+1,1)=tdft(1:m+1,1)+1.d0  
  tdft(:,6)=tb0             
  uv=tb0
  NULLIFY(y)
  y => dft(:,3)
  CALL l_lzj_u(v(1,2),y,ta0,tb0)
  tdft(:,3)=-ta0
  lzu=ta0
  tdft(:,8)=tb0 
  NULLIFY(y)
  y => dft(:,1)
  CALL l_lzj_u(a(1,1),y,ta0,tb0)
  tdft(:,7)=ta0 
  tdft(:,4)=tb0 
  NULLIFY(y)
  y => a(:,2)
  CALL ldft(y,ta0)
  CALL l_uzj(a(1,2),ta0,tb0)
  NULLIFY(y)
  y => tdft(:,5)
  z => tdft(:,2)
  CALL l_lzj_u(a(1,2),ta0,y,z)
  tdft(:,5)=y
  tdft(:,2)=z
  x(1)=a(1,2)
  DO i=2,m
     x(i)=-a(1,2)*v(m-i+2,1)
     x(i)=x(i)-a(1,1)*v(m-i+2,2)
     x(i)=x(i)+v(1,1)*a(m-i+2,2)
     x(i)=x(i)+v(1,2)*a(m-i+2,1)
  END DO
  NULLIFY(y)
  y => s(:,1)
  CALL cgata(dft,ta0,tb0,tdft,x,y,eps)
  w(m:n-1)=x
  s(:,1)=x
  x = v(m:1:-1,1)
  NULLIFY(y)
  y => s(:,2)
  CALL ax(dft,ta0,tb0,x)
  CALL cgaat(dft,ta0,tb0,tdft,x,y,eps)
  w(m-1:1:-1) = x(1:m-1)
  s(:,2)=x
  CALL tpzdft(w)
  t(:,1)=w
  NULLIFY(y)
  y => s(:,3)
  x = a(:,1)
  CALL atx(tdft,x)
  CALL cgata(dft,ta0,tb0,tdft,x,y,eps)
  s(:,3)=x
  CALL ldft(x,w)
  t(:,2)=w 
  NULLIFY(y)
  y => t(:,5)
  CALL l_uzj(x(1),w,y)
  t(:,5)=y 
  x = v(m:1:-1,2)
  NULLIFY(y)
  y => s(:,4)
  CALL ax(dft,ta0,tb0,x)
  CALL cgaat(dft,ta0,tb0,tdft,x,y,eps)
  s(:,4)=x
  x(:)=x(m:1:-1)
  NULLIFY(y)
  NULLIFY(z)
  y => t(:,4)
  z => t(:,3)
  CALL ldft(x,y)
  CALL l_uzj(x(1),y,z)
  t(:,4) = y
  t(:,3) = z
END SUBROUTINE btl_inv

!====================================================================
!                          SUBROUTINE ATAX
!====================================================================
! This subroutine computes the vector y=A^T A x, where A and A^T are
! defined by the DFT dft, u, v:
! A~= u*dft(4)-dft(1)*dft(5)+dft(2)*v+dft(3)*dft(6)
! A^T ~= tdft(1)*tdft(2)+tdft(3)*tdft(4)+tdft(5)*tdft(6)+tdft(7)*tdft(8) 
!====================================================================
! Input variables:
!  dft, u, v, tdft : pointers associated with the DFT of A and A^T
!  x               : pointer associated with the vector x
!====================================================================
! Output variables:
!  y               : pointer associated with the vector y
!====================================================================
SUBROUTINE atax(dft,u,v,tdft,x,y)
  USE bt_interface, ONLY : ax, atx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER :: dft, tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: x,y
  y=x
  CALL ax(dft,u,v,y)
  CALL atx(tdft,y)
END SUBROUTINE atax


!====================================================================
!                          SUBROUTINE AATX
!====================================================================
! This subroutine computes the vector y=AA^T x, where A and A^T are
! defined by the DFT dft, u, v:
! A~= u*dft(4)-dft(1)*dft(5)+dft(2)*v+dft(3)*dft(6)
! A^T ~= tdft(1)*tdft(2)+tdft(3)*tdft(4)+tdft(5)*tdft(6)+tdft(7)*tdft(8) 
!====================================================================
! Input variables:
!  dft, u, v, tdft : pointers associated with the DFT of A and A^T
!  x               : pointer associated with the vector x
!====================================================================
! Output variables:
!  y               : pointer associated with the vector y
!====================================================================
SUBROUTINE aatx(dft,u,v,tdft,x,y)
  USE bt_interface, ONLY : ax, atx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER :: dft, tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: x, y
  y=x
  CALL atx(tdft,y)
  CALL ax(dft,u,v,y)
END SUBROUTINE aatx


!====================================================================
!                          SUBROUTINE AX
!====================================================================
! This subroutine computes the vector y=Ax, where A is
! defined by the DFT dft, u, v:
! A~= u*dft(4)-dft(1)*dft(5)+dft(2)*v+dft(3)*dft(6)
!====================================================================
! Input variables:
!  dft, u, v : pointers associated with the DFT of A 
!  x         : pointer associated with the vector x
!====================================================================
! Output variables:
!  y         : pointer associated with the vector y
!====================================================================
SUBROUTINE ax(dft,u,v,x)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: t, w
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, r
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  IF(.NOT.ASSOCIATED(z))THEN
     ALLOCATE(z(n),s(n),r(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z,s,r)
        ALLOCATE(z(n),s(n),r(n))
     ENDIF
  ENDIF
  s(1:m)=x
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  w => dft(:,4)
  CALL luv(u,w,z,s)
  NULLIFY(w)
  t => dft(:,1)
  w => dft(:,5)
  CALL luv(t,w,z,r)
  s=s-r
  NULLIFY(t)
  t => dft(:,2)
  CALL luv(t,v,z,r)
  s=s+r
  NULLIFY(t)
  NULLIFY(w)
  t => dft(:,3)
  w => dft(:,6)
  CALL luv(t,w,z,r)
  s=s+r
  CALL iffts1(s,r)
  x=r(1:m)
END SUBROUTINE ax


!====================================================================
!                          SUBROUTINE ATX
!====================================================================
! This subroutine computes the vector y=A^T x, where A^T is
! defined by the DFT tdft:
! A^T ~= tdft(1)*tdft(2)+tdft(3)*tdft(4)+tdft(5)*tdft(6)+tdft(7)*tdft(8)
!====================================================================
! Input variables:
!  tdft : pointers associated with the DFT of A^T
!  x    : pointer associated with the vector x
!====================================================================
! Output variables:
!  y    : pointer associated with the vector y
!====================================================================
SUBROUTINE atx(tdft,x)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: t, w
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, r
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  IF(.NOT.ASSOCIATED(z))THEN
     ALLOCATE(z(n),s(n),r(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z,s,r)
        ALLOCATE(z(n),s(n),r(n))
     ENDIF
  ENDIF
  s(1:m)=x
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  t => tdft(:,1)
  w => tdft(:,2)
  CALL luv(t,w,z,s)
  NULLIFY(t)
  NULLIFY(w)
  t => tdft(:,3)
  w => tdft(:,4)
  CALL luv(t,w,z,r)
  s=s+r
  NULLIFY(t)
  NULLIFY(w)  
  t => tdft(:,5)
  w => tdft(:,6)
  CALL luv(t,w,z,r)
  s=s+r
  NULLIFY(t)
  NULLIFY(w)
  t => tdft(:,7)
  w => tdft(:,8)
  CALL luv(t,w,z,r)
  s=s+r
  CALL iffts1(s,r)
  x=r(1:m)
END SUBROUTINE atx


!====================================================================
!                          SUBROUTINE VCR
!====================================================================
! This subroutine updates the vectors v_{-1}, v_0, after one step of
! cyclic reduction.
!====================================================================
! Input variables:
!  v : pointer associated with v_{-1}, v_0
!  t : pointer associated with the DFT of A_0^{-1}
!  dft, thb, tr, ta: pointers associated with the DFT of A_i
!====================================================================
! Output variables: new values for the pointer v
!====================================================================
SUBROUTINE vcr(v,t,dft,thb,tr,ta)
  USE fft_interface
  USE bt_interface, ONLY : a0invv, av1, av2
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: v, t, dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: thb, tr, ta
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, w
  INTEGER                                       :: n, m
  INTRINSIC size
  m=SIZE(v,1)
  n=m+m
  IF(.NOT.ASSOCIATED(s))THEN
     ALLOCATE(s(n),w(n),z(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z,s,w)
        ALLOCATE(s(n),w(n),z(n))
     ENDIF
  ENDIF
  s=dft(:,3)
  CALL a0invv(t,s,z,1)
  CALL av2(dft,thb,tr,ta,z,w)
  CALL iffts1(w,s)
  v(1:m,2)=-s(1:m)
  CALL av1(dft,z,w)
  CALL iffts1(w,s)
  v(1:m,1)=v(1:m,1)-s(1:m)
END SUBROUTINE vcr


!====================================================================
!                          SUBROUTINE ACR
!====================================================================
! This subroutine updates the vectors a_{-1}, a_0, a_1 after one step
! of cyclic reduction.
!====================================================================
! Input variables:
!  a : pointer associated with a_{-1}, a_0, a_1
!  ha, r : pointers associated with the displacement structure of
!  of A_i, \widehat A
!  dft, thb, tr, ta: pointers associated with the DFT of A_i
!====================================================================
! Output variables: new values for a, ha, r
!====================================================================
SUBROUTINE acr(a,ha,r,t,dft,thb,tr,ta)
  USE fft_interface
  USE bt_interface, ONLY : a0invv, av1, av2
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: a, t, dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: ha, r, thb, tr, ta
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, w
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  INTEGER                                       :: n, m
  INTRINSIC size
  m=SIZE(a,1)
  n=m+m
  IF(.NOT.ASSOCIATED(s))THEN
     ALLOCATE(s(n),z(n),w(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z,s,w)
        ALLOCATE(s(n),z(n),w(n))
     ENDIF
  ENDIF
  x => dft(:,1)
  CALL a0invv(t,x,s,1)
  CALL av1(dft,s,w)
  CALL iffts1(w,z)
  a(1:m,1)=-z(1:m)
  CALL av2(dft,thb,tr,ta,s,w)
  CALL iffts1(w,z)
  r(1:m)=r(1:m)-z(1:m)
  a(1:m,2)=a(1:m,2)-z(1:m)
  CALL a0invv(t,ta,s,1)
  CALL av2(dft,thb,tr,ta,s,w)
  CALL iffts1(w,z)
  a(1:m,3)=-z(1:m)
  CALL av1(dft,s,w)
  CALL iffts1(w,z)
  a(1:m,2)=a(1:m,2)-z(1:m)
  ha(1:m)=ha(1:m)-z(1:m)
END SUBROUTINE acr

!====================================================================
!                          SUBROUTINE BCR
!====================================================================
! This subroutine updates the vectors b, after one step
! of cyclic reduction.
!====================================================================
! Input variables:
!  b : pointer associated with b
!  t, dft, thb, tr, ta: pointers associated with the DFT of A_i and
!  A_0^{-1} 
!====================================================================
! Output variables: new values for b and the pointer bev containing 
! T^{-1}b_+
!====================================================================
SUBROUTINE bcr(b,bev,t,dft,thb,tr,ta)
  USE fft_interface
  USE bt_interface, ONLY : a0invv, av1, av2
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER    :: t, dft, b, bev
  REAL(KIND(0.d0)), DIMENSION(:), POINTER      :: thb, tr, ta
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: temp
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, w,y
  INTEGER                                       :: n, m, nb, i
  INTRINSIC size
  m=SIZE(b,1)
  nb=SIZE(b,2)
  n=m+m
  IF(ASSOCIATED(bev))NULLIFY(bev,temp)
  ALLOCATE(bev(m,nb/2),temp(m,nb/2))
  IF(.NOT.ASSOCIATED(s))THEN
     ALLOCATE(s(n),z(n),w(n),y(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z,s,w,y)
        ALLOCATE(s(n),z(n),w(n),y(n))
     ENDIF
  ENDIF
  s(1:m)=b(:,2)
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  CALL a0invv(t,z,w,0)
  bev(:,1)=w(1:m)
  w(m+1:n)=0.d0
  CALL ffts1(w,s)
  CALL av1(dft,s,z)
  CALL iffts1(z,w)
  b(:,1)=b(:,1)-w(1:m)
  IF(nb==2)THEN
     w(1:m)=b(:,1)
     DEALLOCATE(b)
     ALLOCATE(b(m,1))
     b(:,1)=w(1:m)
     RETURN
  ENDIF
  DO i=2,nb/2
     CALL av2(dft,thb,tr,ta,s,y)
     s(1:m)=b(:,2*i)
     s(m+1:n)=0.d0
     CALL ffts1(s,z)
     CALL a0invv(t,z,w,0)
     bev(:,i)=w(1:m)
     w(m+1:n)=0.d0
     CALL ffts1(w,s)
     CALL av1(dft,s,w)
     y=y+w
     CALL iffts1(y,z)
     b(:,i)=b(:,i)-z(1:m)
  END DO
  temp=b(:,1:nb/2)
  DEALLOCATE(b)
  ALLOCATE(b(m,nb/2))
  b=temp(:,1:nb/2)
END SUBROUTINE bcr


!====================================================================
!                          SUBROUTINE A0INVV
!====================================================================
! This subroutine computes the vector A_0^{-1}x, if e/=1, otherwise
! the DFT of the vector A_0^{-1}x.
!====================================================================
! Input variables:
!  t : pointer associated with the DFT of A_0^{-1}
!  v : pointer associated with the DFT of the vector x, of dimension
!  n=2m, where m is the dimension of x
!  e : flag
!====================================================================
! Output variables: the n-dimensional vector w such that:
! - if e=1, w is the DFT associated with A_0^{-1}x;
! - if e/=1, the first m components of w are the vector A_0^{-1}x.
!====================================================================
SUBROUTINE a0invv(t,v,w,e)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: t
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: v, w
  INTEGER                                       :: e
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: z
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: s1, s2
  INTEGER                                       :: n, m
  INTRINSIC size
  n=SIZE(v,1)
  m=n/2
  IF(n/=SIZE(z,1))NULLIFY(z)
  IF(.NOT.ASSOCIATED(z))ALLOCATE(z(n))
  s1 => t(:,1)
  CALL convs(s1,v,w)
  NULLIFY(s1)
  s1 => t(:,2)
  s2 => t(:,3)
  CALL luv(s1,s2,v,z)
  w=w+z
  NULLIFY(s1)
  NULLIFY(s2)
  s1 => t(:,4)
  s2 => t(:,5)
  CALL luv(s1,s2,v,z)
  w=w-z
  CALL iffts1(w,z)
  IF(e==1)THEN
     z(m+1:n)=0.d0
     CALL ffts1(z,w)
  ELSE
     w(1:m)=z(1:m)
  ENDIF
END SUBROUTINE a0invv


!====================================================================
!                          SUBROUTINE AV1
!====================================================================
! This subroutine computes the the convolution w= A_{-1}*s, where s is
! the DFT of a vector
!====================================================================
! Input variables:
!  s : pointer associated with the DFT of a vector
!  dft : pointer associated with the DFTs of matrix A_{-1}
!====================================================================
! Output variables: 
! the vector w=dft1*(dft4*s)+dft2*(dft6*s)
!====================================================================
SUBROUTINE av1(dft,s,w)
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: s, w
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x1, x2
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: z
  INTEGER                                       ::  n
  INTRINSIC size
  n=SIZE(s,1)
  IF(.NOT.ASSOCIATED(z)) THEN
     ALLOCATE(z(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z)
        ALLOCATE(z(n))
     ENDIF
  ENDIF
  x1 => dft(:,1)
  x2 => dft(:,4)
  CALL luv(x1,x2,s,w)
  NULLIFY(x1)
  NULLIFY(x2)
  x1 => dft(:,2)
  x2 => dft(:,6)
  CALL luv(x1,x2,s,z)
  w=w+z
END SUBROUTINE av1

!====================================================================
!                          SUBROUTINE AV2
!====================================================================
! This subroutine computes the the convolution w= A_1*s, where s is
! the DFT of a vector
!====================================================================
! Input variables:
!  s : pointer associated with the DFT of a vector
!  dft, thb, tr, ta : pointers associated with the DFTs of matrix A_1
!====================================================================
! Output variables: 
! the vector w=ta*s-tr*(dft5*s)+dft3*(thb*s)
!====================================================================
SUBROUTINE av2(dft,thb,tr,ta,s,w)
  USE bt_interface, ONLY : luv
  USE fft_interface
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: thb, tr, ta, s, w
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: z
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  INTEGER                                       :: n
  INTRINSIC size
  n=SIZE(s,1)
  IF(.NOT.ASSOCIATED(z)) THEN
     ALLOCATE(z(n))
  ELSE
     IF(n/=SIZE(z,1))THEN
        NULLIFY(z)
        ALLOCATE(z(n))
     ENDIF
  ENDIF
  CALL convs(ta,s,w)
  x => dft(:,5)
  CALL luv(tr,x,s,z)
  w=w-z
  NULLIFY(x)
  x => dft(:,3)
  CALL luv(x,thb,s,z)
  w=w+z
END SUBROUTINE av2

!====================================================================
!                          SUBROUTINE LUV
!====================================================================
! This subroutine computes the the convolution z=a*(b*s)
!====================================================================
! Input variables:
!  s : pointer associated with the DFT of a vector
!  a : pointer associated with the DFT of a lower triangular Toeplitz
! matrix 
!  b : pointer associated with the DFT of an upper triangular Toeplitz
! matrix
!====================================================================
! Output variables: 
! the vector z=a*(b*s)
!====================================================================
SUBROUTINE luv(a,b,s,z)
  USE fft_interface
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: a, b, s, z
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: w
  INTEGER                                       :: n, m
  INTRINSIC size
  n=SIZE(a,1)
  m=n/2
  IF(.NOT.ASSOCIATED(w)) THEN
     ALLOCATE(w(n))
  ELSE
     IF(n/=SIZE(w,1))THEN
        NULLIFY(w)
        ALLOCATE(w(n))
     ENDIF
  ENDIF
  CALL convs(b,s,w)
  CALL iffts1(w,z)
  z(m+1:n)=0.d0
  CALL ffts1(z,w)
  CALL convs(a,w,z)
END SUBROUTINE luv


!====================================================================
!                          SUBROUTINE CGATA
!====================================================================
! This subroutine computes the solution x of the system A^T A x= b by
! means of the CG method
!====================================================================
! Input variables:
! dft, u, v, tdft: pointers associated with the DFTs of the matrices
! A, A^T, i.e.:
!  A~= u*dft(4)-dft(1)*dft(5)+dft(2)*v+dft(3)*dft(6)
!  A^T ~= tdft(1)*tdft(2)+tdft(3)*tdft(4)+tdft(5)*tdft(6)+
!         tdft(7)*tdft(8) 
!  x  : pointer associated with the vector b
!  x0 : pointer associated with the starting approximation for the CG
! method 
!  eps: error bound for the stopping condition
!====================================================================
! Output variables: 
! x   : the solution
!====================================================================
SUBROUTINE cgata(dft,u,v,tdft,x,x0,eps)
  USE bt_interface, ONLY : atax
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft, tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x, x0
  REAL(KIND(0.d0))                              :: eps
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: y, r, xp
  REAL(KIND(0.d0))                              :: alpha, beta, err,&
       & eps1, rn, rn1
  INTEGER                                       :: m, n, iter, &
       & maxiter 
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  maxiter=5*m
  IF(.NOT.ASSOCIATED(y))THEN
     ALLOCATE(y(m),r(m),xp(m))
  ELSE
     IF(m/=SIZE(y,1))THEN
        NULLIFY(y,r,xp)
        ALLOCATE(y(m),r(m),xp(m))
     ENDIF
  ENDIF
  rn1=DOT_PRODUCT(x,x)
  eps1=eps*SQRT(rn1)
  CALL atax(dft,u,v,tdft,x0,y)
  r=x-y
  rn1=DOT_PRODUCT(r,r)
  err=SQRT(rn1)
  x=x0
  iter=0
  cg: DO WHILE(err>eps1)
     iter=iter+1
     IF(iter>maxiter)THEN
        WRITE(*,*)"Reached the maximum number of iterations"
        STOP
     ENDIF
     IF (iter==1) THEN
        beta=0.d0
        xp=r
     ELSE
        beta=rn1/rn
        xp(:)=r(:)+beta*xp(:)
     ENDIF
     CALL atax(dft,u,v,tdft,xp,y)
     alpha=DOT_PRODUCT(xp,y)
     alpha=rn1/alpha
     x(:)=x(:)+alpha*xp(:)
     r(:)=r(:)-alpha*y(:)
     rn=rn1
     rn1=DOT_PRODUCT(r,r)
     err=SQRT(rn1)
  END DO cg
END SUBROUTINE cgata



!====================================================================
!                          SUBROUTINE CGAAT
!====================================================================
! This subroutine computes the solution x of the system AA^T x= b by
! means of the CG method
!====================================================================
! Input variables:
! dft, u, v, tdft: pointers associated with the DFTs of the matrices
! A, A^T, i.e.:
!  A~= u*dft(4)-dft(1)*dft(5)+dft(2)*v+dft(3)*dft(6)
!  A^T ~= tdft(1)*tdft(2)+tdft(3)*tdft(4)+tdft(5)*tdft(6)+
!         tdft(7)*tdft(8) 
!  x  : pointer associated with the vector b
!  x0 : pointer associated with the starting approximation for the CG
! method 
!  eps: error bound for the stopping condition
!====================================================================
! Output variables: 
! x   : the solution
!====================================================================
SUBROUTINE cgaat(dft,u,v,tdft,x,x0,eps)
  USE bt_interface, ONLY : ax, aatx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft, tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x, x0
  REAL(KIND(0.d0))                              :: eps
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: y, r, xp
  REAL(KIND(0.d0))                              :: alpha, beta, err,&
       & eps1, rn, rn1
  INTEGER                                       :: m, n, iter, &
       & maxiter 
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  maxiter=5*m
  IF(.NOT.ASSOCIATED(y))THEN
     ALLOCATE(y(m),r(m),xp(m))
  ELSE
     IF(m/=SIZE(y,1))THEN
        NULLIFY(y,r,xp)
        ALLOCATE(y(m),r(m),xp(m))
     ENDIF
  ENDIF
  rn1=DOT_PRODUCT(x,x)
  eps1=eps*SQRT(rn1)
  CALL aatx(dft,u,v,tdft,x0,y)
  r=x-y
  rn1=DOT_PRODUCT(r,r)
  err=SQRT(rn1)
  x=x0
  iter=0
  cg: DO WHILE(err>eps1)
     iter=iter+1
     IF(iter>maxiter)THEN
        WRITE(*,*)"Reached the maximum number of iterations"
        STOP
     ENDIF
     IF (iter==1) THEN
        beta=0.d0
        xp=r
     ELSE
        beta=rn1/rn
        xp(:)=r(:)+beta*xp(:)
     ENDIF
     CALL aatx(dft,u,v,tdft,xp,y)
     alpha=DOT_PRODUCT(xp,y)
     alpha=rn1/alpha
     x(:)=x(:)+alpha*xp(:)
     r(:)=r(:)-alpha*y(:)
     rn=rn1
     rn1=DOT_PRODUCT(r,r)
     err=SQRT(rn1)
  END DO cg
END SUBROUTINE cgaat


!====================================================================
!                          SUBROUTINE STORE
!====================================================================
! This subroutine stores the pointers dft, tr, ta, thb, bev, t for
! the back substitutiuon stage
!====================================================================
SUBROUTINE store(dft,tr,ta,thb,bev,t)
  USE accumulo
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: tr, ta, thb
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: dft, bev, t
  INTEGER :: n,m,ln,nb
  INTRINSIC size
  m=SIZE(bev,1)
  nb=SIZE(bev,2)
  n=m+m
  IF(SIZE(dfts,1)/=n)NULLIFY(dfts,ts,trs,tas,thbs,bevs)
  IF(.NOT.ASSOCIATED(dfts))THEN
     ln=LOG(nb*1.d0)/LOG(2.d0)
     IF(2**ln<nb)ln=ln+1
     ln=ln+1
     ALLOCATE(dfts(n,6,ln),ts(n,5,ln),trs(n,ln),tas(n,ln),thbs(n,ln)&
          & ,bevs(m,2*nb)) 
     ib=0
     iter=1
  ENDIF
  trs(:,iter)=tr
  tas(:,iter)=ta
  thbs(:,iter)=thb
  dfts(:,:,iter)=dft
  ts(:,:,iter)=t
  bevs(:,ib+1:ib+nb)=bev
  iter=iter+1
  ib=ib+nb
END SUBROUTINE store

!====================================================================
!                          SUBROUTINE RECOVER
!====================================================================
! This subroutine recovers, by means of back substitutiuon, the
! solution x of the system Ax=b
!====================================================================
SUBROUTINE recover(conv,bm0,ha,r1,dft,tr,ta,thb,t,b,x,eps)
  USE comtr
  USE accumulo
  USE bt_interface, ONLY : solve1, solve2, solvel
  IMPLICIT NONE
  REAL(KIND(0.d0))                                  :: r1, eps
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: bm0, ha
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: b, x
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: ta, tr, thb
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: dft, bev, t
  LOGICAL                                           :: conv
  INTEGER                                           :: nb, i
  nb=SIZE(b,2)
  iter=iter-1
  IF(conv) THEN
     CALL solve1(ha,bm0,r1,tr,thb,dft,t,iter,b,x,eps)
  ELSE
     CALL solvel(ha,bm0,ta,thb,dft,b,x,eps)
  END IF
  ib=ib-nb
  DO i=iter,1,-1
     NULLIFY(ta,tr,thb,dft,t,bev)
     ta => tas(:,i)
     tr => trs(:,i)
     thb => thbs(:,i)
     dft => dfts(:,:,i)
     t => ts(:,:,i)
     bev => bevs(:,ib+1:ib+nb)
     CALL solve2(ta,tr,thb,dft,t,bev,i,x)
     nb=nb*2
     ib=ib-nb
  END DO
END SUBROUTINE recover

!====================================================================
!                          SUBROUTINE SOLVEL
!====================================================================
! This subroutine computes the solution x of the latter system
! generated by cyclic reduction, in the case where the convergence is
! reached (and the system is bidiagonal)
!====================================================================
! Input variables:
! ha, bm0, ta, thb, dft: pointers associated with the DFTs of the
! limit bidiagonal matrix.
!  b    : pointer associated with the vector b
!  eps: error bound for the CG method for inverting the diagonal
! matrix 
!====================================================================
! Output variables: 
!  xtot : pointer associated with the solution of the initial system
!====================================================================
SUBROUTINE solvel(ha,bm0,ta,thb,dft,b,xtot,eps)
  USE comtr
  USE fft_interface
  USE tdft_interface
  USE bt_interface, ONLY : hathalx, halx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: ha, bm0, ta, thb, x
  REAL(KIND(0.d0))                              :: eps
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: dft
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: tha, that, xtot,b
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: y, r, xp,s,w
  REAL(KIND(0.d0))                              :: alpha, beta, err,&
       & eps1, rn, rn1
  INTEGER                                       :: m, n, iter, maxiter 
  INTRINSIC size
  m=SIZE(xtot,1)
  n=m+m
  maxiter=5*m
  ALLOCATE(y(m),r(m),xp(m),x(m),w(n),s(n),tha(n,5),that(n,5))
  w(1:m)=ha
  w(m+1)=0.d0
  w(m+2:n)=bm0(m-1:1:-1)
  CALL ffts1(w,s)
  tha(:,1)=s
  CALL ttrasp(s,w)
  that(:,1)=w
  tha(:,2)=dft(:,1)
  tha(:,3)=dft(:,5)
  tha(:,4)=dft(:,2)
  tha(:,5)=thb
  that(:,2)=lzu
  CALL ttrasp(ta,s)
  that(:,3)=s
  CALL ttrasp(thb,s)
  that(:,4)=s
  that(:,5)=uv
  x=b(:,1)
  CALL halx(that,x)
  rn1=DOT_PRODUCT(x,x)
  err=SQRT(rn1)
  eps1=eps*err
  r=x
  iter=0
  cg: DO WHILE(err>eps1)
     iter=iter+1
     IF(iter>maxiter)THEN
        WRITE(*,*)"Reached the maximum number of iterations"
        STOP
     ENDIF
     IF (iter==1) THEN
        beta=0.d0
        xp=x
        x=0.d0
     ELSE
        beta=rn1/rn
        xp(:)=r(:)+beta*xp(:)
     ENDIF
     CALL hathalx(tha,that,xp,y)
     alpha=DOT_PRODUCT(xp,y)
     alpha=rn1/alpha
     x(:)=x(:)+alpha*xp(:)
     r(:)=r(:)-alpha*y(:)
     rn=rn1
     rn1=DOT_PRODUCT(r,r)
     err=SQRT(rn1)
  END DO cg
  xtot(:,1)=x
END SUBROUTINE solvel


!====================================================================
!                          SUBROUTINE HATHALX
!====================================================================
! This subroutine computes the the product y = \hat A^T \hat A x
!====================================================================
! Input variables:
! tha, that: pointers associated with the DFTs of the matrices \hat A
! and \hat A^T
! x: pointer associated with the vector x
!====================================================================
! Output variables: 
! y: pointer associated with the vector y
!====================================================================
SUBROUTINE hathalx(tha,that,x,y)
  USE bt_interface, ONLY : halx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER :: tha,that
  REAL(KIND(0.d0)), DIMENSION(:), POINTER     :: x,y
  y=x
  CALL halx(tha,y)
  CALL halx(that,y)
END SUBROUTINE hathalx


!====================================================================
!                          SUBROUTINE HALX
!====================================================================
! This subroutine computes the the product y = \hat A x
!====================================================================
! Input variables:
! tha: pointer associated with the DFTs of the matrices \hat A
! x: pointer associated with the vector x
!====================================================================
! Output variables: 
! y: pointer associated with the vector y
!====================================================================
SUBROUTINE halx(tha,x)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: tha
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, r, z
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: v, w
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  IF(SIZE(s,1)/=n)NULLIFY(s,r)
  IF(.NOT.ASSOCIATED(s))ALLOCATE(s(n),r(n),z(n))
  s(1:m)=x
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  v=>tha(:,1)
  CALL convs(v,z,s)
  NULLIFY(v)
  v=>tha(:,2)
  w=>tha(:,3)
  CALL luv(v,w,z,r)
  s=s-r
  NULLIFY(v)
  NULLIFY(w)
  v=>tha(:,4)
  w=>tha(:,5)
  CALL luv(v,w,z,r)
  s=s+r
  CALL iffts1(s,r)
  x=r(1:m)
END SUBROUTINE halx


!====================================================================
!                          SUBROUTINE SOLVE1
!====================================================================
! This subroutine computes the solution x of the latter system
! generated by cyclic reduction, in the case where there is not
! convergence 
!====================================================================
! Input variables:
! ha, bm0, r1, tr,thb, dft, t: pointers associated with the DFTs of
! the latter matrix 
! limit bidiagonal matrix.
!  b    : pointer associated with the vector b
!  eps: error bound for the CG method for inverting the diagonal
! matrix 
!====================================================================
! Output variables: 
!  xtot : pointer associated with the solution of the initial system
!====================================================================
SUBROUTINE solve1(ha,bm0,r1,tr,thb,dft,t,iter,b,x,eps)
  USE comtr
  USE bt_interface, ONLY : a0invv, av1, ha_solve, hatx, hax
  USE fft_interface
  USE tdft_interface
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: ha, bm0, tr, thb
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: dft,t,  x, b
  REAL(KIND(0.d0))                                  :: r1
  INTEGER                                           :: iter
! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: z, y, s, w, tv
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: tdft
  REAL(KIND(0.d0))                                  :: eps
  INTEGER                                           :: j, m, n, i, nb
  nb=SIZE(b,2)
  m=SIZE(ha,1)
  n=m+m
  ALLOCATE(y(m))
  ALLOCATE(z(n))
  ALLOCATE(s(n))
  ALLOCATE(w(n))
  ALLOCATE(tdft(n,3))
  j=2**iter
  w(1:m)=b(:,nb)
  w(m+1:n)=0.d0
  CALL ffts1(w,s)
  CALL a0invv(t,s,z,0)
  x(:,1+j*(nb-1))=z(1:m)
  DO i=nb-2,1,-1
     z(m+1:n)=0.d0
     CALL ffts1(z,s)
     CALL av1(dft,s,w)
     CALL iffts1(w,s)
     s(m+1:n)=0.d0
     s(1:m)=-s(1:m)+b(1:m,i+1)
     CALL ffts1(s,w)
     CALL a0invv(t,w,z,0)
     x(:,1+j*i)=z(1:m)
  END DO
  z(m+1:n)=0.d0
  CALL ffts1(z,s)
  CALL av1(dft,s,w)
  CALL iffts1(w,s)
  s(1:m) = -s(1:m)+b(1:m,1)
  y=s(1:m)
  w(1:m)=ha
  w(m+1)=0.d0
  w(m+2:n)=bm0(m-1:1:-1)
  CALL ffts1(w,s)
  tv => dft(:,2)
  CALL ttrasp(s,w)
  tdft(:,1)=w
  CALL l_lzj(r1,tr,w)
  tdft(:,2)=w
  tdft(:,3)=uv
  CALL hatx(tdft,y)
  CALL ha_solve(s,tv,thb,tdft,y,eps)
  x(:,1)=y
END SUBROUTINE solve1

!====================================================================
!                          SUBROUTINE SOLVE2
!====================================================================
! This subroutine computes by back substitution the solution x of the
! initial system  
!====================================================================
! Input variables:
! ta,tr,thb,dft,t,bev: pointers associated with the DFTs of
! the matrices involved in the back substitution;
! iter: integer
!====================================================================
! Output variables: 
!  x : pointer associated with the solution of the initial system
!====================================================================
SUBROUTINE solve2(ta,tr,thb,dft,t,bev,iter,x)
  USE bt_interface, ONLY : a0invv, av1, av2
  USE fft_interface
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: ta, tr, thb
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER         :: dft,t, bev, x
  INTEGER                                           :: iter
! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: z, s, w
  INTEGER                                           :: j, m, n, i, nb
  nb=SIZE(bev,2)
  m=SIZE(x,1)
  n=m+m
  ALLOCATE(z(n))
  ALLOCATE(s(n))
  ALLOCATE(w(n))
  j=2**(iter-1)
  w(1:m)=x(:,1+2*j*(nb-1))
  w(m+1:n)=0.d0
  CALL ffts1(w,s)
  CALL av2(dft,thb,tr,ta,s,w)
  CALL iffts1(w,z)
  z(m+1:n)=0.d0
  CALL ffts1(z,w)
  CALL a0invv(t,w,z,0)
  x(:,1+j*(2*nb-1))=bev(:,nb)-z(1:m)
  DO i=nb-2,0,-1
     CALL av1(dft,s,w)
     z(1:m)=x(:,1+2*j*i)
     z(m+1:n)=0.d0
     CALL ffts1(z,s)
     CALL av2(dft,thb,tr,ta,s,z)
     w=w+z
     CALL iffts1(w,z)
     z(m+1:n)=0.d0
     CALL ffts1(z,w)
     CALL a0invv(t,w,z,0)
     x(:,1+j*(2*i+1))=bev(:,i+1)-z(1:m)
  END DO
END SUBROUTINE solve2


!====================================================================
!                          SUBROUTINE HATHAX
!====================================================================
! This subroutine computes the product y = \hat A^T \hat A x
!====================================================================
! Input variables:
! ta, tv, tb, tdft: pointers associated with the DFTs of the matrices
! \hat A and \hat A^T
! x: pointer associated with the vector x
!====================================================================
! Output variables: 
! y: pointer associated with the vector y
!====================================================================
SUBROUTINE hathax(ta,tv,tb,tdft,x,y)
  USE bt_interface, ONLY : hax, hatx
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER :: tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: ta, tv, tb
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: x,y
  y=x
  CALL hax(ta,tv,tb,y)
  CALL hatx(tdft,y)
END SUBROUTINE hathax


!====================================================================
!                          SUBROUTINE HAX
!====================================================================
! This subroutine computes the product y = \hat A x
!====================================================================
! Input variables:
! ta, tv, tb: pointers associated with the DFTs of the matrix \hat A
! x: pointer associated with the vector x
!====================================================================
! Output variables: 
! y: pointer associated with the vector y
!====================================================================
SUBROUTINE hax(ta,tv,tb,x)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: ta, tv, tb
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, r, z
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  IF(SIZE(s,1)/=n)NULLIFY(s,r,z)
  IF(.NOT.ASSOCIATED(s))ALLOCATE(s(n),r(n),z(n))
  s(1:m)=x
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  CALL convs(ta,z,s)
  CALL luv(tv,tb,z,r)
  s=s+r
  CALL iffts1(s,r)
  x=r(1:m)
END SUBROUTINE hax

!====================================================================
!                          SUBROUTINE HATX
!====================================================================
! This subroutine computes the product y = \hat A^T  x
!====================================================================
! Input variables:
! tdft: pointer associated with the DFTs of the matrices \hat A^T
! x: pointer associated with the vector x
!====================================================================
! Output variables: 
! y: pointer associated with the vector y
!====================================================================
SUBROUTINE hatx(tdft,x)
  USE fft_interface
  USE bt_interface, ONLY : luv
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: t, w
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: s, z, r
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  IF(SIZE(s,1)/=n)NULLIFY(s,r,z)
  IF(.NOT.ASSOCIATED(z))ALLOCATE(z(n),s(n),r(n))
  s(1:m)=x
  s(m+1:n)=0.d0
  CALL ffts1(s,z)
  t => tdft(:,1)
  CALL convs(t,z,s)
  NULLIFY(t)
  t => tdft(:,2)
  w => tdft(:,3)
  CALL luv(t,w,z,r)
  s=s+r
  CALL iffts1(s,r)
  x=r(1:m)
END SUBROUTINE hatx


!====================================================================
!                          SUBROUTINE HA_SOLVE
!====================================================================
! This subroutine computes the solution x of the system A^T A x= b
! by means of the CG method
!====================================================================
! Input variables:
! ta,tv,tb,tdft: pointers associated with the DFTs of the matrices
! A, A^T, i.e.:
!  A~= ta+tv*tb
!  A^T ~= tdft(1)+tdft(2)*tdft(3)
!  x  : pointer associated with the vector b
!  eps: error bound for the stopping condition
!====================================================================
! Output variables: 
! x   : the solution
!====================================================================
SUBROUTINE ha_solve(ta,tv,tb,tdft,x,eps)
  USE bt_interface, ONLY : hathax
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER     :: tdft
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: ta, tv, tb
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: x
  REAL(KIND(0.d0))                              :: eps
  ! Auxiliary variables
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: y, r, xp
  REAL(KIND(0.d0))                              :: alpha, beta, err,&
       & eps1, rn, rn1
  INTEGER                                       :: m, n, iter, maxiter 
  INTRINSIC size
  m=SIZE(x,1)
  n=m+m
  maxiter=5*m
  IF(SIZE(y,1)/=m)NULLIFY(y,r,xp)
  IF(.NOT.ASSOCIATED(y))ALLOCATE(y(m),r(m),xp(m))
  rn1=DOT_PRODUCT(x,x)
  err=SQRT(rn1)
  eps1=eps*err
  r=x
  iter=0
  cg: DO WHILE(err>eps1)
     iter=iter+1
     IF(iter>maxiter)THEN
        WRITE(*,*)"Reached the maximum number of iterations"
        STOP
     ENDIF
     IF (iter==1) THEN
        beta=0.d0
        xp=x
        x=0.d0
     ELSE
        beta=rn1/rn
        xp(:)=r(:)+beta*xp(:)
     ENDIF
     CALL hathax(ta,tv,tb,tdft,xp,y)
     alpha=DOT_PRODUCT(xp,y)
     alpha=rn1/alpha
     x(:)=x(:)+alpha*xp(:)
     r(:)=r(:)-alpha*y(:)
     rn=rn1
     rn1=DOT_PRODUCT(r,r)
     err=SQRT(rn1)
  END DO cg
END SUBROUTINE ha_solve

!====================================================================
!                          SUBROUTINE TEST
!====================================================================
! This subroutine performes the test for the stopping condition in
! cyclic reduction method
!====================================================================
! Input variables:
! a, v, r: pointers associated with the matrices A_i
!====================================================================
! Output variables: 
! err: the error bound
!====================================================================
SUBROUTINE test(a,v,r,err)
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER :: a, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER   :: r
  REAL(KIND(0.d0))                          :: err
  err=SUM(ABS(a(:,3)))+SUM(ABS(r(:)))*SUM(ABS(v(:,2)))
END SUBROUTINE test

