!====================================================================
! file fft_bt.f90
!====================================================================
!         EFFECTIVE METHODS FOR SOLVING BANDED TOEPLITZ SYSTEMS
!                     by D.A. Bini and B. Meini
!               (bini@dm.unipi.it    meini@dm.unipi.it)
!           Fortran 90 Program version 1.0, October 30, 1997
!--------------------------------------------------------------------
!                   S C A L A R        C A S E
!====================================================================
!                           FFT SUBROUTINES 
!====================================================================
! This file contains a set of subroutines needed in order
! to perform computations involving the Discrete Fourier Transform
! DFT). More precisely the following subroutines are reported:    
!                                                                 
! fillroots : compute the roots of 1                               
! ifft1     : compute the inverse DFT                              
! fft1      : compute the DFT                                       
! iffts1    : compute the inverse DFT of the DFT of a real vector 
! ffts1     : compute the DFT of a real vector 
! tpzdft    : compute the DFT associated with a Toeplitz matrix
! ldft      : compute the DFT associated with a lower triangular
! Toeplitz matrix
! convs     : compute convolution
! l_uzj     : compute DFT
! l_lzj_u   : compute DFT
! l_lzj     : compute DFT
! ttrasp    : compute DFT
!====================================================================

!====================================================================
!                      SUBROUTINE FILLROOTS
!====================================================================
! Compute the n-th roots of 1 if they have not been yet computed
! before, otherwise return.
! The real and the imaginary parts of the roots are stored in the
! vectors wr, wi, respectively. These vectors are put in a common block.
!====================================================================
SUBROUTINE fillroots(n)
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER           :: wr, wi
  REAL(KIND(0.d0)), DIMENSION(:), ALLOCATABLE, SAVE :: wwr, wwi
  REAL(KIND(0.d0))                                  :: pi, pi2
  INTEGER                                           :: i, j, k, m, mi1, n
  COMMON wr, wi
  INTRINSIC size, allocated
  pi= 6.28318530717958647692528676656d0/n
  k=SIZE(wr,1)
  IF(.NOT.ALLOCATED(wwr))THEN
     ALLOCATE(wwr(n))
     ALLOCATE(wwi(n))
     ALLOCATE(wr(n))
     ALLOCATE(wi(n))
     loop1 : DO i=1,n
	pi2=(i-1)*pi
	wwr(i)=COS(pi2)
	wwi(i)=SIN(pi2)
     END DO loop1
     wr=wwr
     wi=wwi
     RETURN
  END IF
  IF(n<=k)THEN
     RETURN
  END IF
  DEALLOCATE(wr)
  DEALLOCATE(wi)
  ALLOCATE(wr(n))
  ALLOCATE(wi(n))
  m=n/k
  loop2 : DO i=1,k
     mi1=m*(i-1)
     wr(mi1+1)=wwr(i)
     wi(mi1+1)=wwi(i)
     loop3 : DO j=2,m
	wr(mi1+j)=COS(pi*(mi1+j-1))
	wi(mi1+j)=SIN(pi*(mi1+j-1))
     END DO loop3
  END DO loop2
  DEALLOCATE(wwr)
  DEALLOCATE(wwi)
  ALLOCATE(wwr(n))
  ALLOCATE(wwi(n))
  wwr=wr
  wwi=wi
  RETURN
END SUBROUTINE fillroots




!========================================================================!
!                      SUBROUTINE IFFT1                                  !
!========================================================================!
! Compute the Inverse Discrete Fourier Transform  of the vector having   !
! real parts stored in the vector x and imaginary parts stored in the    !
! vector y. On output x and y contain the real and the imaginary parts   !
! of the transformed vector, respectively.                               !
! The algorithm is an adaptation of the split-radix FFT by               !
! Dhuamel-Hollman (Sorensen et al. IEEE trans.                           !
! on Acoustics Speech and Signal Processing, ASSP-34, 1986).             !
!========================================================================!
SUBROUTINE  ifft1(x,y) 
  USE fft_interface, ONLY : fillroots
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: x,y
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: wr,wi
  REAL(KIND(0.d0))                         :: r1, r2, s1, s2, s3, ss1, &
       ss3, cc1,  cc3, xt, yt, un
  INTEGER                                  :: i, j, k, m, n, nmax, n2, &
       n4, ne, na, is, id,      &
       i0, i1, i2, i3, n1 ,na3
  COMMON wr,wi 
  INTRINSIC size
  n=SIZE(x,1)
  m=LOG(n*1.d0)/LOG(2.d0)
  IF(2**m<n)m=m+1
  n2=2*n
  CALL fillroots(n)
  nmax=SIZE(wr,1)
  loop1 : DO  k=1,m-1
     n2=n2/2
     n4=n2/4
     ne=nmax/n2  +1
     na=1
     loop2 : DO j=1,n4
	na3=3*(na-1)+1
	cc1=wr(na)
	ss1=wi(na)
	cc3=wr(na3)
	ss3=wi(na3)
	na=j*(ne-1)+1
	is=j
	id=2*n2
40	loop3 : DO i0=is,n-1,id
	   i1=i0+n4
	   i2=i1+n4
	   i3=i2+n4
	   r1=x(i0)-x(i2)
	   x(i0)=x(i0)+x(i2)
	   r2=x(i1)-x(i3)
	   x(i1)=x(i1)+x(i3)
	   s1=y(i0)-y(i2)
	   y(i0)=y(i0)+y(i2)
	   s2=y(i1)-y(i3)
	   y(i1)=y(i1)+y(i3)
	   s3=r1-s2
	   r1=r1+s2
	   s2=r2-s1
	   r2=r2+s1
	   x(i2)=r1*cc1-s2*ss1
	   y(i2)=-s2*cc1-r1*ss1
	   x(i3)=s3*cc3+r2*ss3
	   y(i3)=r2*cc3-s3*ss3
	END DO loop3
	is=2*id-n2+j
	id=4*id
	IF (is<n) GOTO 40
     END DO loop2
  END DO loop1
  is=1
  id=4
50 loop6 : DO  i0=is,n,id
     i1=i0+1
     r1=x(i0)
     x(i0)=r1+x(i1)
     x(i1)=r1-x(i1)
     r1=y(i0)
     y(i0)=r1+y(i1)
     y(i1)=r1-y(i1)
  END DO loop6
  is=2*id-1
  id=4*id
  IF(is<n)GOTO 50
  ! ----DIGIT REVERSE COUNTER  -----------------
  j=1
  n1=n-1
  loop14 : DO  i=1,n1
     IF(i>=j)GOTO 101
     xt=x(j)
     yt=y(j)
     y(j)=y(i)
     y(i)=yt
     x(j)=x(i)
     x(i)=xt
101  k=n/2
102  IF (k>=j)GOTO 103
     j=j-k
     k=k/2
     GOTO 102
103  j=j+k
  END DO loop14
  un=1.d0/n
  x=x*un
  y=y*un
  RETURN
END SUBROUTINE ifft1



!========================================================================!
!                      SUBROUTINE FFT1                                   !
!========================================================================!
! Compute the Discrete Fourier Transform  of the vector having           !
! real parts stored in the vector x and imaginary parts stored in the    !
! vector y. On output x and y contain the real and the imaginary parts   !
! of the transformed vector, respectively.                               !
! The algorithm is an adaptation of the split-radix FFT by               !
! Dhuamel-Hollman (Sorensen et al. IEEE trans.                           !
! on Acoustics Speech and Signal Processing, ASSP-34, 1986).             !
!========================================================================!
SUBROUTINE  fft1(x,y) 
  USE fft_interface, ONLY : fillroots
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: x,y
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: wr,wi
  REAL(KIND(0.d0))                         :: r1, r2, r3, s2, ss1, ss3, &
       cc1, cc3, xt, yt, s1
  INTEGER                                  :: i, j, k, m, n, nmax, n1,  &
       n2, n4, ne, na, is, id,   &
       i0, i1, i2, i3, na3
  COMMON wr,wi 
  INTRINSIC size
  n=SIZE(x,1)
  m=LOG(n*1.d0)/LOG(2.d0)
  IF(2**m<n)m=m+1
  n2=2*n
  CALL fillroots(n)
  nmax=SIZE(wr,1)
  ! -------------DIGIT REVERSE COUNTER ---------------
  j=1
  n1=n-1
  loop104: DO i=1,n1
     IF(i>=j)GOTO 101
     xt=x(j)
     yt=y(j)
     y(j)=y(i)
     y(i)=yt
     x(j)=x(i)
     x(i)=xt
101  k=n/2
102  IF (k>=j)GOTO 103
     j=j-k
     k=k/2
     GOTO 102
103  j=j+k
  END DO loop104
  ! ---------------LENGTH TWO BUTTERFLIES--------------
  is=1
  id=4
70 loop60 : DO i0=is,n,id
     i1=i0+1
     r1=x(i0)
     x(i0)=r1+x(i1)
     x(i1)=r1-x(i1)
     r1=-y(i0)
     y(i0)=-r1+y(i1)
     y(i1)=-r1-y(i1)
  END DO loop60
  is=2*id-1
  id=4*id
  IF (is<n) GOTO 70
  !----------------------------------------------------
  n2=2 
  loop10 : DO  k=2,m
     n2=2*n2
     n4=n2/4
     ne=nmax/n2
     na=1
     loop20 : DO  j=1,n4
	na3=3*(na-1)+1
	cc1=wr(na)
	ss1=wi(na)
	cc3=wr(na3)
	ss3=wi(na3)
	na=j*ne+1
	is=j
	id=2*n2
40	loop30 : DO  i0=is,n-1,id
	   i1=i0+n4
	   i2=i1+n4
	   i3=i2+n4
	   r1=x(i2)*cc1-y(i2)*ss1
	   s1=-y(i2)*cc1-x(i2)*ss1
	   r2=x(i3)*cc3-y(i3)*ss3
	   s2=-y(i3)*cc3-x(i3)*ss3
	   r3=r1+r2
	   r2=r1-r2
	   r1=s1+s2
	   s2=s1-s2
	   x(i2)=x(i0)-r3
	   x(i0)=x(i0)+r3
	   x(i3)=x(i1)-s2
	   x(i1)=x(i1)+s2
	   y(i2)=y(i0)+r1
	   y(i0)=y(i0)-r1
	   y(i3)=y(i1)-r2
	   y(i1)=y(i1)+r2
	END DO loop30
        is=2*id-n2+j
        id=4*id
        IF(is<n)GOTO 40
     END DO loop20
  END DO loop10
  RETURN
END SUBROUTINE fft1



!========================================================================!
!                      SUBROUTINE FFTS1                                  !
!========================================================================!
! Compute the Discrete Fourier Transform (DFT) of the real vector u      !
! of size n, by means of an fft of half the size.                        !
! On output the vector v contains the components of the transform        !
! in such a way that v(j)+I*v(n-j+1) is the j-th component of the        !
! transformed vector, j=2,...,n/2, where I**2=-1. Moreover v(1) and      !
! v(n/2+1) coincide with  the first and the (n/2+1)-st components        !
! of the transformed vector, respectively.                               !
!========================================================================!
SUBROUTINE ffts1(u,v)
  USE fft_interface, ONLY : fillroots,fft1
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: wr, wi
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: zr, zi
  REAL(KIND(0.d0))                         :: uq, x, y
  INTEGER                                  :: i, l, km, n, ln, nmax
  COMMON wr,wi 
  INTRINSIC size
  n=SIZE(u,1)
  ln=LOG(n*1.d0)/LOG(2.d0)
  IF(2**ln<n)ln=ln+1
  CALL fillroots(n)
  nmax=SIZE(wr,1)
  km=LOG(nmax*1.d0)/LOG(2.d0)
  IF(2**km<nmax)km=km+1
  IF(.NOT.ASSOCIATED(zr))THEN
     ALLOCATE(zr(n/2))
     ALLOCATE(zi(n/2))
  ELSE
     IF(SIZE(zr,1)/=n/2)THEN
        NULLIFY(zr)
        NULLIFY(zi)
        ALLOCATE(zr(n/2))
        ALLOCATE(zi(n/2))
     ENDIF
  ENDIF
  loop20 : DO l=0,n/2-1
     zr(l+1)=u(2*l+1)
     zi(l+1)=u(2*l+2)
  END DO loop20
  CALL fft1(zr,zi)
  v(1)=zr(1)+zi(1)
  v(n/2+1)=zr(1)-zi(1)
  uq=1.d0/2.d0
  loop30 : DO  l=2,n/2
     i=2**(km-ln)*(l-1)+1
     x=wr(i)*(zr(l)-zr(n/2-l+2))
     y=wi(i)*(zi(l)+zi(n/2-l+2))
     v(n-l+2)=(zi(l)-zi(n/2-l+2)-x+y)*uq
     x=wr(i)*(zi(l)+zi(n/2-l+2))
     y=wi(i)*(zr(l)-zr(n/2-l+2))
     v(l)=(zr(l)+zr(n/2-l+2)+x+y)*uq
  END DO loop30
  RETURN
END SUBROUTINE ffts1



!========================================================================!
!                      SUBROUTINE IFFTS1                                 !
!========================================================================!
! Compute the Inverse Discrete Fourier Transform (IDFT) of the vector    !
! u of size n which is the DFT of a real vector, by means of an ifft of  !
! half the size.                                                         !
! The input vector u is such that u(j)+I*u(n-j+1) is the j-th component  !
! of the vector to be transformed, j=2,...,n/2. Moreover u(1) and        !
! u(n/2+1) coincide with the first and the (n/2+1)-st components,        !
! respectively.                                                          !
! On output the vector v contains the (real) components of the IDFT      !
! This subroutine provides the solution of Problem 2.                    !
!========================================================================!
SUBROUTINE iffts1(u,v)
  USE fft_interface, ONLY : fillroots, ifft1
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: u, v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER  :: wr, wi
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE:: zr, zi
  REAL(KIND(0.d0))                         :: dr
  INTEGER                                  :: l, km, n, ln, nmax, nm, &
       k1, k2, mm, ml
  COMMON wr,wi 
  INTRINSIC size
  n=SIZE(u,1)
  ln=LOG(n*1.d0)/LOG(2.d0)
  IF(2**ln<n)ln=ln+1
  CALL fillroots(n)
  nmax=SIZE(wr,1)
  km=LOG(nmax*1.d0)/LOG(2.d0)
  IF(2**km<nmax)km=km+1
  IF(.NOT.ASSOCIATED(zr))THEN
     ALLOCATE(zr(n/2))
     ALLOCATE(zi(n/2))
  ELSE
     IF(SIZE(zr,1)/=n/2)THEN
        NULLIFY(zr)
        NULLIFY(zi)
        ALLOCATE(zr(n/2))
        ALLOCATE(zi(n/2))
     ENDIF
  ENDIF
  mm=2**(km-ln)
  nm=n/2
  dr=1.0d0/2.0d0
  loop20 : DO  l=1,nm-1
     ml=l*mm+1
     k1=l+1
     k2=k1+nm
     zr(k1)=u(k1)+u(nm-l+1)+wr(ml)*(-u(nm+k1)-u(n-l+1))
     zr(k1)=(zr(k1)+wi(ml)*(-u(nm-l+1)+u(k1)))*dr
     zi(k1)=-u(nm+k1)+u(n-l+1)+wr(ml)*(u(k1)-u(nm-l+1))
     zi(k1)=(zi(k1)+wi(ml)*(u(nm+k1)+u(n-l+1)))*dr
  END DO loop20
  zr(1)=(u(1)+u(nm+1))*dr
  zi(1)=(u(1)-u(nm+1))*dr
  CALL ifft1(zr,zi)
  loop30 : DO  l=1,nm
     v(l*2)=zi(l)
     v(2*l-1)=zr(l)
  END DO loop30
  RETURN
END SUBROUTINE iffts1



!====================================================================
!                      SUBROUTINE TPZDFT                        
!====================================================================
! Computes the DFT of the vector 
! t_n t_{n+1} ... t_{2n-1} 0 t_1 t_2 ... t_{n-1}
! that defines the Toeplitz matrix
!  t_n     t_{n-1} ... t_1
!  t_{n+1} t_n         t_2
!  ......................
!  t_{2n-1}            t_n
!====================================================================
SUBROUTINE tpzdft(t)
  USE fft_interface, ONLY: ffts1
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: t
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: a
  INTEGER                                       :: m, n
  INTRINSIC size
  m=SIZE(t,1)
  n=m/2
  IF(.NOT.ASSOCIATED(a))THEN
     ALLOCATE(a(m))
  ELSE
     IF(SIZE(a,1)/=m)THEN
        NULLIFY(a)
        ALLOCATE(a(m))
     ENDIF
  ENDIF
  a(1:n)=t(n:m-1)
  a(n+1)=0.d0
  a(n+2:m)=t(1:n-1)
  CALL ffts1(a,t)
END SUBROUTINE tpzdft


!====================================================================
!                      SUBROUTINE LDFT                        
!====================================================================
! Computes the DFT of the vector 
! low_1 low_2 ... low_n 0 ..... 0
! that defines the Toeplitz matrix
!  low_1     0 .... 0
!  low_2  low_1 0   0
!  ......................
!  low_n .....    low_1
!====================================================================
SUBROUTINE ldft(low,tlow)
  USE fft_interface, ONLY : ffts1
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: low, tlow
  REAL(KIND(0.d0)), DIMENSION(:), POINTER, SAVE :: a
  INTEGER                                       :: n, m
  INTRINSIC size
  n=SIZE(low,1)
  m=n+n
  IF(.NOT.ASSOCIATED(a))THEN
     ALLOCATE(a(m))
  ELSE
     IF(SIZE(a,1)/=m)THEN
        NULLIFY(a)
        ALLOCATE(a(m))
     ENDIF
  ENDIF
  a(1:n)=low(1:n)
  a(n+1:m)=0.d0
  CALL ffts1(a,tlow)
END SUBROUTINE ldft


!====================================================================
!                      SUBROUTINE CONVS                     
!====================================================================
! Compute the convolution between the vectors u and v, which are DFTs
! of real vectors
!====================================================================
SUBROUTINE convs(u,v,a)
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:), POINTER       :: u, v, a
  REAL(KIND(0.d0))                              :: z, w
  INTEGER                                       :: n, nm, j1, j2
  INTRINSIC size
  n=SIZE(u,1)
  nm=n/2
  a(1)=u(1)*v(1)
  a(nm+1)=u(nm+1)*v(nm+1)
  cmplx : DO j1=2,nm
     j2=n-j1+2
     z=u(j1)*v(j1)
     w=u(j2)*v(j2)
     a(j1)=z-w
     a(j2)=-z-w
     z=u(j1)+u(j2)
     w=v(j1)+v(j2)
     a(j2)=a(j2)+z*w
  END DO cmplx
END SUBROUTINE convs


!====================================================================
!                      SUBROUTINE L_UZJ                   
!====================================================================
! Compute the DFT tw associated with  L^T(ZJw), given w and the DFT
! associated with L(w)
!====================================================================
SUBROUTINE l_uzj(v,w,tw)
  IMPLICIT NONE
  REAL(KIND(0.d0))                        :: v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER :: w, tw
  INTEGER :: n, m
  n=SIZE(tw,1)
  m=n/2+1
  tw(1:m)=w(1:m)-v
  tw(m+1:n)=w(m+1:n)
  tw(2:n:2)=-tw(2:n:2)
END  SUBROUTINE l_uzj

!====================================================================
!                      SUBROUTINE L_LZJ_U                   
!====================================================================
! Compute the DFT tw1 associated with  L(ZJw), and the DFT tw2
! associated with L^T(w), given w_1 and the DFT associated with L(w) 
!====================================================================
SUBROUTINE l_lzj_u(v,w,tw1,tw2)
  IMPLICIT NONE
  REAL(KIND(0.d0))                        :: v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER :: w, tw1, tw2
  INTEGER                                 :: n, m
  n=SIZE(tw1,1)
  m=n/2+1
  tw2(m+1:n)=-w(m+1:n)
  tw2(1:m)=w(1:m)
  tw1=tw2
  tw1(1:m)=tw1(1:m)-v
  tw1(2:n:2)=-tw1(2:n:2)
END  SUBROUTINE l_lzj_u

!====================================================================
!                      SUBROUTINE L_LZJ                  
!====================================================================
! Compute the DFT tw1 associated with  L(ZJw), 
! given w_1 and the DFT associated with L(w) 
!====================================================================
SUBROUTINE l_lzj(v,w,tw1)
  IMPLICIT NONE
  REAL(KIND(0.d0))                        :: v
  REAL(KIND(0.d0)), DIMENSION(:), POINTER :: w, tw1
  INTEGER                                 :: n, m
  n=SIZE(tw1,1)
  m=n/2+1
  tw1(m+1:n)=-w(m+1:n)
  tw1(1:m)=w(1:m)
  tw1(1:m)=tw1(1:m)-v
  tw1(2:n:2)=-tw1(2:n:2)
END  SUBROUTINE l_lzj

!====================================================================
!                      SUBROUTINE TTRASP                
!====================================================================
! Compute the DFT w associated with  L^T(x), 
! given the DFT associated with L(x) 
!====================================================================
SUBROUTINE ttrasp(s,w)
  IMPLICIT NONE
  INTEGER                                 :: n, m
  REAL(KIND(0.d0)), DIMENSION(:), POINTER :: s, w
  n=SIZE(w,1)
  m=n/2+1
  w(1:m)=s(1:m)
  w(m+1:n)=-s(m+1:n)
END SUBROUTINE ttrasp






























































